class FakeRequest:
    ''' Very simple Request Simulator that will allow for testing 
    Mixins without a full http cycle '''
    def __init__(self, GET=None, POST=None, META=None, method=None, path_info=''):
        self.GET = GET if GET else {}
        self.POST = POST if POST else {}
        self.META = META if META else {}
        self.method = method
        self.path_info = path_info


class FakeFormsetFactory:
    def __call__(self, data=None, instance=None):
        self.data = data
        self.instance = instance
        return self


import pandas as pd
from pandas import DataFrame as DF

from django.db import models
from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType


class Country(models.Model):
    ''' A instance of this class represents a country in the world using the ISO 3166-1 standard '''
    id = models.BigAutoField(primary_key=True)
    alpha2 = models.CharField(max_length=2, null=True)
    alpha3 = models.CharField(max_length=3, unique=True)
    num = models.PositiveIntegerField(unique=True)
    name = models.CharField(max_length=250)

    @classmethod
    def create_objs(cls, update=False):
        url = 'https://www.iban.com/country-codes'
        df = pd.read_html(url)[0]
        df.columns = ['name', 'alpha2', 'alpha3', 'num']
        
        for tup in df.itertuples():
            alpha2 = None if pd.isna(tup.alpha2) else tup.alpha2
            goc_method = cls.objects.updated_or_create if update else cls.objects.get_or_create
            goc_method(
                alpha3 = tup.alpha3,
                defaults = dict(
                    alpha2=alpha2, num=tup.num, name=tup.name
                )
            )

    def __str__(self):
        return f'{self.alpha3}'


class Subdivision(models.Model):
    ''' Represents a country dub division using the ISO 3166-2 standard '''
    id = models.BigAutoField(primary_key=True)
    country = models.ForeignKey('Country', on_delete=models.PROTECT)
    iso_code = models.CharField(max_length=10, unique=True)
    name = models.CharField(max_length=250)
    category = models.CharField(max_length=50)

    
    @staticmethod
    def clean_df(df, country):
        MANUAL_CATEGORY = {
            'AF':'Province'
        }
        if 'Code' not in df.columns:
            return DF()
        rtups = [('Code','code')]
        for col in df.columns:
            if 'subdivision name (en)' in col.lower():
                rtups.append((col, 'name'))
            if 'subdivision category' in col.lower():
                rtups.append((col, 'category'))
        keep_cols = dict(rtups).values()
        for rtup in rtups:
            df.columns = df.columns.str.replace(*rtup, regex=False)      
        df = df.copy()[keep_cols]
        if 'name' not in df.columns:
            df['name'] = 'N/A'
        if 'category' not in df.columns:
            df['category'] = MANUAL_CATEGORY.get(country.alpha2, 'subdivision')
        return df


    @classmethod
    def get_create_df(cls, country):
        """ method retreives the data needed to create objects from wikipedia """
        url = f'https://en.wikipedia.org/wiki/ISO_3166-2:{country.alpha2}'
        dfs = pd.read_html(url)
        df = DF()
        for tdf in dfs:
            tdf = cls.clean_df(tdf, country)
            if tdf.empty: continue
            df = pd.concat([df, tdf])
        df.drop_duplicates('code', keep='first', inplace=True)
        return df


    @classmethod
    def create_objs_from_country(cls, country):
        ''' Method takes a coutnry object and creates all Subdivisions for that country '''
        df = cls.get_create_df(country)
        if df.empty: 
            return
        for tup in df.itertuples():
            cls.objects.get_or_create(
                iso_code = tup.code,
                defaults = dict(
                    name = tup.name, category = tup.category, country=country
                )
            )

    def __str__(self):
        return f'{self.iso_code} | {self.name}'


class Coordinate(models.Model):
    id = models.BigAutoField(primary_key=True)
    latitude = models.DecimalField(max_digits=8, decimal_places=6)
    longitude = models.DecimalField(max_digits=9, decimal_places=6)

    class Meta:
        constraints = [
            models.UniqueConstraint(fields=['latitude', 'longitude'], name='unique_coord')
        ]


class Address(models.Model):
    id = models.BigAutoField(primary_key=True)
    name = models.CharField(max_length=255, null=True, blank=True)
    attn = models.CharField(max_length=255, null=True, blank=True)
    street1 = models.CharField(max_length=510, null=True, blank=True)
    street2 = models.CharField(max_length=510, null=True, blank=True)
    street3 = models.CharField(max_length=510, null=True, blank=True)
    city = models.CharField(max_length=255, null=True, blank=True)
    subdivision = models.ForeignKey('django_aux_geo.Subdivision', on_delete=models.PROTECT, null=True, blank=True)
    postal_code = models.CharField(max_length=15, null=True, blank=True)
    country = models.ForeignKey('django_aux_geo.Country', on_delete=models.PROTECT, null=True, blank=True)
    coordinate = models.ForeignKey('django_aux_geo.Coordinate', on_delete=models.SET_NULL, null=True, blank=True)

    # Allow for generic FK realtionships to Address
    content_type = models.ForeignKey(ContentType, on_delete=models.CASCADE, null=True, blank=True)
    object_id = models.CharField(max_length=255, null=True, blank=True)
    fk_object = GenericForeignKey("content_type", "object_id")


    def __str__(self):
        return f'''
        Name: {self.name}
        ATTN: {self.attn}
        Street1: {self.street1}
        Street2: {self.street2}
        Street3: {self.street3}
        City: {self.city}
        Subdivision: {self.subdivision}
        Postal Code: {self.postal_code}
        Country: {self.country}
        '''


from django.test import TestCase
from django_aux_geo.models import *

#------------GEO TESTS------------




class TestCountry(TestCase):
    ''' TestCase for the Country model '''

    def test_create_objs(self):
        Country.create_objs()
        for alpha2 in ['AD','CA','FR','US']:
            self.assertTrue(Country.objects.filter(alpha2=alpha2).exists())


from django.test import TestCase, RequestFactory, Client
from django_aux.models import *
from django_aux.views import *
from django_aux.utils import PasswordUtils
from .models import Person
from .views import PersonLookup, PersonCreate, PersonCreateWithRequest
from .filters import PersonFilter
from django.contrib.auth.models import User
from django.views.generic import *
from django import forms
from django.test.client import RequestFactory
from django.test import Client
from django.urls import path, reverse
from django.contrib.sessions.middleware import SessionMiddleware
from .utils import *

#------------Django-Aux TESTS------------

class TestPasswordUtils(TestCase):
    ''' Test Case for PasswordUtils '''

    def test_check_password(self):
        raise_tests = [
            dict(password='<10', min_length=10, must_have_symbol=False, must_have_caps=False, must_have_digit=False),
            dict(password='password', min_length=1, must_have_symbol=True, must_have_caps=False, must_have_digit=False),
            dict(password='password', min_length=1, must_have_symbol=False, must_have_caps=True, must_have_digit=False),
            dict(password='password', min_length=1, must_have_symbol=False, must_have_caps=False, must_have_digit=True),
        ]
        for test in raise_tests:
            self.assertRaises(AssertionError, PasswordUtils().check_password, test)

        ok_tests = [
            dict(password='password', min_length=1, must_have_symbol=False, must_have_caps=False, must_have_digit=False),
            dict(password='password!', min_length=1, must_have_symbol=True, must_have_caps=False, must_have_digit=False),
            dict(password='passworD', min_length=1, must_have_symbol=False, must_have_caps=True, must_have_digit=False),
            dict(password='password1', min_length=1, must_have_symbol=False, must_have_caps=False, must_have_digit=True),
        ]
        for test in ok_tests:
            # Just make sure there are no raises
            PasswordUtils().check_password(**test)

    def test_generate_password(self):
        for i in range(5,50, 5):
            pw = PasswordUtils().generate_password(N=i)
            # Ensure Length is correct
            self.assertEqual(len(pw), i)
            # Ensure no amgigous characters
            self.assertEqual(len(set(pw).intersection(PasswordUtils.AMBIGUOUS_CHARS)), 0)


class TestSaveFilterMixin(TestCase):
    ''' Test Case for SaveFilterMixin '''

    def setUp(self):
        # create a request factory that all tests have access to
        self.factory = RequestFactory()
        self.client = Client()
        self.user = User.objects.create_user('billy', password = 'password')
        self.client.login(username='billy', password='password')
        

    def test_savefiltermixin_get(self):
        response = self.client.get('/person-lookup')
        self.assertIsNone(response.wsgi_request.session.get('PersonLookup_qstr'))
        response = self.client.get('/person-lookup', {'last_name__icontains': 'hyatt', 'first_name__icontains':'jordan'})
        qstr = response.wsgi_request.session.get('PersonLookup_qstr')
        self.assertEqual(qstr, 'last_name__icontains=hyatt&first_name__icontains=jordan')
        response = self.client.get('/person-lookup')
        qstr = response.wsgi_request.session.get('PersonLookup_qstr')
        self.assertEqual(qstr, 'last_name__icontains=hyatt&first_name__icontains=jordan')
        response = self.client.get('/person-lookup', {'clear_filter': 'Clear Filter'})
        qstr = response.wsgi_request.session.get('PersonLookup_qstr')
    
    def test_clear_filter(self):
        request = self.factory.get('/person-lookup', {'last_name__icontains': 'hyatt', 'first_name__icontains':'jordan'})
        request.user = self.user
        view = PersonLookup()
        view.setup(request)
        kwargs = view.get_filterset_kwargs(PersonFilter)
        self.assertNotEqual(kwargs, {})
        request = self.factory.get('/person-lookup', {'last_name__icontains': 'hyatt', 'clear_filter': 'Clear Filter'})
        request.user = self.user
        view = PersonLookup()
        view.setup(request)
        kwargs = view.get_filterset_kwargs(PersonFilter)
        self.assertEqual(kwargs, {})

class TestRedirectPrevMixin(TestCase):
    '''A test for the RedirectPrevMixin view mixin '''

    def setUp(self):
        #Make access to RequestFactory
        self.factory = RequestFactory()
        self.user = User.objects.create_user(
            username='billy', email='bill@bob.com', password='foobar')
        self.client = Client()
        self.client.login(username='billy', password='foobar')
        self.middleware = SessionMiddleware(lambda x: None)

    def test_form_takes_request_arg(self):
        ''' Test the form_takes_request_arg property method '''
        self.assertFalse(PersonCreate().form_takes_request_arg)
        self.assertTrue(PersonCreateWithRequest().form_takes_request_arg)


    def test_get_form_kwargs(self):
        ''' Test the get_forms_kwargs method of a view using RedirectPrevMixin '''
        request = self.factory.get('/person-create')
        self.assertNotIn('request', PersonCreate(request=request).get_form_kwargs().keys())
        request = self.factory.get('/person-create-request')
        self.assertIn('request', PersonCreateWithRequest(request=request).get_form_kwargs().keys())


    def test_get_next_is_exception(self):
        request = self.factory.get('/person-create')
        pc = PersonCreate(request=request)
        reverse_name = pc.get_next_is_exception('person-delete')
        self.assertEqual(reverse_name, 'person-lookup')

        request = self.factory.get('/person-create')
        pc = PersonCreate(request=PersonCreateWithRequest)
        reverse_name = pc.get_next_is_exception('this-is-ok')
        self.assertEqual(reverse_name, None)


    def test_next(self):
        response = self.client.get('/person-create', HTTP_REFERER='final-destination')
        self.assertEqual(response.wsgi_request.session.get('next'), 'final-destination')

        response = self.client.get('/person-create', HTTP_REFERER='person-delete')
        self.assertEqual(response.wsgi_request.session.get('next'), '/person-lookup')



class TestInlineFormsetMixin(TestCase):
    
    def test_get_current_extra(self):
        ifm = InlineFormsetMixin()
        ifm.max_extra = 10
        ifm.request = FakeRequest(GET = {'extra':None})
        self.assertEqual(ifm.get_current_extra(), 1)
        ifm.request = FakeRequest(GET = {'extra':1})
        self.assertEqual(ifm.get_current_extra(), 1)
        ifm.request = FakeRequest(GET = {'extra':2})
        self.assertEqual(ifm.get_current_extra(), 2)
        # try one that exceeds max
        ifm.request = FakeRequest(GET = {'extra':11})
        self.assertEqual(ifm.get_current_extra(), 10) 

    def test_add_factories_to_context(self):
        ifm = InlineFormsetMixin()
        ifm.factories = [
            {'factory':FakeFormsetFactory(), 'helper':None, 'header':None},
            {'factory':FakeFormsetFactory(), 'helper':None, 'header':None},
        ]
        ifm.request = FakeRequest(method='GET')
        ifm.object = None
        context = {}
        ifm.add_factories_to_context(context)
        num_factories = len(context.get('factories'))
        self.assertEqual(num_factories, 2)


    def test_set_extra_on_factories(self):
        ifm = InlineFormsetMixin()
        ifm.object = None
        ifm.request = FakeRequest(GET = {'extra':2}, method='GET')
        ifm.factories = [
            {'factory':FakeFormsetFactory(), 'helper':None, 'header':None},
            {'factory':FakeFormsetFactory(), 'helper':None, 'header':None},
        ]
        context = {}
        ifm.add_factories_to_context(context)
        ifm.set_extra_on_factories(context)
        for fd in context.get('factories'):
            self.assertEqual(fd.get('factory').extra, 2)

    def test_add_addlines_url_to_context(self):
        ifm = InlineFormsetMixin()
        tests = (
            {'extra':None, 'target':2},
            {'extra':1, 'target':2},
            {'extra':5, 'target':6},
            {'extra':10, 'target':11},
        )
        for td in tests:
            ifm.request = FakeRequest(GET = {'extra':td.get('extra')})
            context = {}
            ifm.add_addlines_url_to_context(context)
            self.assertEqual(context['addlines_url'], f"?extra={td.get('target')}")


    def test_add_removelines_url_to_context(self):
        ifm = InlineFormsetMixin()
        tests = (
            {'extra':None, 'target':1},
            {'extra':1, 'target':1},
            {'extra':5, 'target':4},
            {'extra':10, 'target':9},
        )
        for td in tests:
            ifm.request = FakeRequest(GET = {'extra':td.get('extra')})
            context = {}
            ifm.add_removelines_url_to_context(context)
            self.assertEqual(context['removelines_url'], f"?extra={td.get('target')}")



def logtest(logger_method):
    def wrap(test_method):
        def inner(instance):
            logger_method(f'Running {test_method.__name__}')
            test_method(instance)
        return inner
    return wrap


def logtestclass(logger_method):
    def wrap(TestClass):
        logger_method(f'Imported {TestClass.__name__}')
        return TestClass
    return wrap



import os
from pathlib import Path


SECRET_KEY = 'fake-test-key'

INSTALLED_APPS = [
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django.contrib.staticfiles',
    'crispy_forms',
    'bootstrap4',
    'bootstrap_datepicker_plus',
    'django_extensions',
    'django_tables2',
    "tests", "django_aux_timeperiods", "django_aux_geo", "django_aux"
]


name = os.getenv('DB_NAME', 'aux_test')
user = os.getenv('DB_USER', 'postgres')
host = os.getenv('DB_HOST', 'localhost')
password = os.getenv('DB_PASSWORD')
port = os.getenv('DB_PORT', '5432')

DATABASES = {
    'default': {
        'ENGINE':'django.db.backends.postgresql',
        'NAME': name,
        'USER': user,
        'PASSWORD': password,
        'HOST': host,
        'PORT': port,
    },
}

DEFAULT_AUTO_FIELD = 'django.db.models.BigAutoField'

MIDDLEWARE = [
    'django.middleware.security.SecurityMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
]

ROOT_URLCONF = 'tests.urls'

BASE_DIR = Path(__file__).resolve().parent.parent

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [os.path.join(BASE_DIR, 'templates')],
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
                'django.contrib.messages.context_processors.messages',
            ],
        },
    },
]



from django.test import TestCase
from django_aux_timeperiods.models import *
import pandas as pd
import datetime as dt
import itertools

#------------TIME PERIOD TESTS------------
class CommonTimePeriodSetup(TestCase):
    ''' Basic setup that runs prior to each Test '''
    
    def setUp(self):
        self.models = [Year, Month, Week, Day]

class TestTimePeriodBase(CommonTimePeriodSetup):
    ''' A Test Class for the TimePeriodBase model '''

    def test_period(self):
        ''' test TimePeriods period property method '''
        #I'm not sure how to test this without just re-running the exact same logic in the method itself.
        #This is import though as all the other tests seem to be using this property method
        today = dt.datetime.now().date()
        testday = dt.datetime(1988,8,28).date() #A great day in History
        for date in [today, testday, dt.datetime(2099,12,31)]:
            day = Day(date=date)
            day.save()
            year,_ = Year.objects.get_or_create(year_num=date.year)
            month,_ = Month.objects.get_or_create(year_num=date.year,month_num=date.month)
            week,_ = Week.objects.get_or_create(year_num=date.year,week_num=date.isocalendar()[1])
            self.assertEqual(pd.Period(date, freq='D'), day.period)            
            self.assertEqual(pd.Period(date, freq='W'), week.period)
            self.assertEqual(pd.Period(year=date.year, freq='Y'), year.period)
            self.assertEqual(pd.Period(year=date.year, month=date.month, freq='M'), month.period)

    def test_get_or_create_n_from_current(self):
        ''' test TimePeriods get_or_create_n_from_current method '''
        now = dt.datetime.now()
        def get_target_period(n, freq):
            now = dt.datetime.now()
            if n == 0:
                return pd.Period(now, freq=freq)
            if n < 0:
                pr = pd.period_range(end = now, periods=abs(n)+1, freq=freq)
                return pr[0]
            else:
                pr = pd.period_range(start = now, periods=abs(n)+1, freq=freq)
                return pr[-1]     

        ns = list(range(-13,15))
        models = [Year, Month, Week, Day]
        for n, model in itertools.product(ns, models):
            obj,_ = model.get_or_create_n_from_current(n=n)
            freq = model.freq_map.get(model.__name__)
            target = get_target_period(n, freq)
            self.assertEqual(obj.period, target)


    def test_get_or_create_from_date(self):
        ''' Test accuracy of the get_or_create_from_date class method '''
        testday = dt.datetime(1988,8,28).date() #A great day in History
        today = dt.datetime.now().date()
        for day in [testday, today]:
            yyyy,_ = Year.get_or_create_from_date(day)
            mm,_ = Month.get_or_create_from_date(day)
            ww,_ = Week.get_or_create_from_date(day)
            dd,_ = Day.get_or_create_from_date(day)
            self.assertEqual(
                yyyy.period, pd.Period(year=day.year, freq='Y')
            )
            self.assertEqual(
                mm.period, pd.Period(year=day.year, month=day.month, freq='M')
            )
            self.assertEqual(
                ww.period, pd.Period(day, freq='W')
            )
            self.assertEqual(
                ww.period.week, day.isocalendar()[1]
            )
            self.assertEqual(
                dd.period, pd.Period(day, freq='D')
            )

    def test_get_or_create_current_period(self):
        ''' Test accuracy of the get_or_create_current_period class method '''
        for cls in self.models:            
            self.assertEqual(
                pd.Period.now(cls.freq_map[cls.__name__]),
                cls.get_or_create_current_period()[0].period
            )

    def test_get_rel_status(self):
        ''' Test Timeperiods get_rel_status method '''    
        now = dt.datetime.now()
        past = now - dt.timedelta(days=366)
        future = now + dt.timedelta(days=366)
        for cls in self.models:
            current,_ = cls.get_or_create_current_period()
            self.assertEqual(current.get_rel_status(), 'present')
            self.assertEqual(current.get_rel_status(dtg=future), 'past')
            self.assertEqual(current.get_rel_status(dtg=past), 'future')

class TestMonth(TestCase):
    ''' A Test Class for the Month model '''

    def test_set_year(self):
        month,_ = Month.get_or_create_current_period()
        self.assertEqual(pd.Period.now('Y'), month.year.period)

class TestWeek(TestCase):
    ''' A Test Class for the Week model '''

    def test_set_year(self):
        week,_ = Week.get_or_create_current_period()
        self.assertEqual(pd.Period.now('Y'), week.year.period)

class TestDay(TestCase):
    ''' A Test Class for the Day model '''

    def test_set_month(self):
        day,_ = Day.get_or_create_current_period()
        self.assertEqual(pd.Period.now('M'), day.month.period)

    def test_set_week(self):
        day,_ = Day.get_or_create_current_period()
        self.assertEqual(pd.Period.now('W'), day.week.period)



from django_aux.filters import FilterSetBase, MetaBase
from crispy_forms.layout import Layout, Fieldset, Row, Div

from .models import Country, Subdivision, Address


class CountryFilter(FilterSetBase):
    class Meta(MetaBase):
        model = Country
        fields = {
            'alpha2': ['exact'],
            'alpha3': ['exact'],
            'name': ['icontains'],

        }
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.form.helper.layout = Layout(
            Fieldset('',
                Row(
                    Div('alpha2', css_class='ml-2 col-flex'),
                    Div('alpha3', css_class='ml-2 col-flex'),
                    Div('name__icontains', css_class='ml-2 col-flex'),
                ),
            ),
        ) 





# Generated by Django 4.0.6 on 2023-03-03 14:03

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Country',
            fields=[
                ('id', models.BigAutoField(primary_key=True, serialize=False)),
                ('alpha2', models.CharField(max_length=2, null=True)),
                ('alpha3', models.CharField(max_length=3, unique=True)),
                ('num', models.PositiveIntegerField(unique=True)),
                ('name', models.CharField(max_length=250)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='Subdivision',
            fields=[
                ('id', models.BigAutoField(primary_key=True, serialize=False)),
                ('iso_code', models.CharField(max_length=10, unique=True)),
                ('name', models.CharField(max_length=250)),
                ('category', models.CharField(max_length=50)),
                ('country', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to='django_aux_geo.country')),
            ],
            options={
                'abstract': False,
            },
        ),
    ]



# Generated by Django 4.0.6 on 2024-02-20 16:34

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
        ('django_aux_geo', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='Address',
            fields=[
                ('id', models.BigAutoField(primary_key=True, serialize=False)),
                ('name', models.CharField(blank=True, max_length=255, null=True)),
                ('attn', models.CharField(blank=True, max_length=255, null=True)),
                ('street1', models.CharField(blank=True, max_length=510, null=True)),
                ('street2', models.CharField(blank=True, max_length=510, null=True)),
                ('street3', models.CharField(blank=True, max_length=510, null=True)),
                ('city', models.CharField(blank=True, max_length=255, null=True)),
                ('postal_code', models.CharField(blank=True, max_length=15, null=True)),
                ('object_id', models.CharField(blank=True, max_length=255, null=True)),
            ],
        ),
        migrations.CreateModel(
            name='Coordinate',
            fields=[
                ('id', models.BigAutoField(primary_key=True, serialize=False)),
                ('latitude', models.DecimalField(decimal_places=6, max_digits=8)),
                ('longitude', models.DecimalField(decimal_places=6, max_digits=9)),
            ],
        ),
        migrations.AddConstraint(
            model_name='coordinate',
            constraint=models.UniqueConstraint(fields=('latitude', 'longitude'), name='unique_coord'),
        ),
        migrations.AddField(
            model_name='address',
            name='content_type',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='contenttypes.contenttype'),
        ),
        migrations.AddField(
            model_name='address',
            name='coordinate',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to='django_aux_geo.coordinate'),
        ),
        migrations.AddField(
            model_name='address',
            name='country',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.PROTECT, to='django_aux_geo.country'),
        ),
        migrations.AddField(
            model_name='address',
            name='subdivision',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.PROTECT, to='django_aux_geo.subdivision'),
        ),
    ]



from django.conf import settings
from django.shortcuts import redirect


class CheckPasswordChangeMiddleware:
    '''
        Middleware checks the user instance to see if a password change is required,
        if so it redirects to the password change path specificed in settings
    '''
    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        fpc_attr = getattr(settings, 'FORCE_PASSWORD_CHANGE_ATTR', 'force_password_change')
        fpc = getattr(request.user, fpc_attr, False)
        is_auth = getattr(request.user,'is_authenticated', False)
        no_pmatch = request.path.rstrip('/') != settings.PASSWORD_CHANGE_URL.rstrip('/')
        if fpc and is_auth and no_pmatch:
            return redirect(settings.PASSWORD_CHANGE_URL)
        return self.get_response(request)


from django import template

register = template.Library()


@register.filter(name='has_group')
def has_group(user, group_name):
    ''' A custom template tag for determining group permissions in templates '''
    return user.groups.filter(name=group_name).exists()


@register.filter
def has_rows(table):
    ''' A custom template tag that returns true if the django table has any entries '''
    rows = getattr(table, 'rows', [])
    return True if len(rows) > 0 else False

@register.filter
def has_items(container):
    ''' A custom template tag that takes an iterable returns true if len > 0 '''
    return True if len(container) > 0 else False

@register.filter
def render_class_name(obj):
    ''' A custom template tag that returns the object's class' name '''
    return obj.__class__.__name__



@register.filter(name='null_default')
def null_default(val, default):
    ''' A custom template tag that will return a default if val is null '''
    return val if val else str(default)



from django.db import connection, ProgrammingError
from django.conf import settings
import string
import random
import logging
from pandas.core.dtypes.dtypes import DatetimeTZDtype
from zoneinfo import ZoneInfo
logger = logging.getLogger(__name__)


class PasswordUtils:
    AMBIGUOUS_CHARS = ['0', 'O', 'l', 'I', 'o']

    def __init__(self, valid_symbols=None):
        if valid_symbols == None:
            self.valid_symbols = ['!','$','#','^']
        else: 
            self.valid_symbols = valid_symbols

    def check_password(
        self, password, min_length=12, must_have_symbol=True, must_have_caps=True, must_have_digit=True
    ):
        pset = set(password)
        assert len(password) >= min_length, f"Password must be at least {min_length} characters"
        assert len(pset.intersection(string.ascii_lowercase)) > 0, f"Password must contain at least 1 lowercase letter"
        if must_have_symbol:
            assert len(pset.intersection(self.valid_symbols)) > 0, f"Password must contain at least 1 special character {self.valid_symbols}"
        if must_have_caps:
            assert len(pset.intersection(string.ascii_uppercase)) > 0, f"Password must contain at least 1 uppercase letter"
        if must_have_digit:
            assert len(pset.intersection(string.digits)) > 0, f"Password must contain at least 1 digit {string.digits}"
        return True


    def generate_password(self, N=15, include_symbols=True, remove_ambiguous=True):
        all = string.ascii_letters + string.digits 
        if include_symbols:
            all += ''.join(self.valid_symbols) * 3
        if remove_ambiguous:
            for bad in self.AMBIGUOUS_CHARS:
                all = all.replace(bad,'')
        return ''.join(random.sample(all, N))


def df_tz_convert(df, timezone=None, format=True, format_str='%a %d %b %Y, %I:%M%p %Z'):
    timezone = timezone or getattr(settings, 'TIME_ZONE')
    if not timezone:
        return df
    tz_obj = ZoneInfo(timezone)
    dtg_cols = df.dtypes[df.dtypes.map(lambda val: isinstance(val, DatetimeTZDtype))].index
    for col in dtg_cols: 
        df[col] = df[col].dt.tz_convert(tz_obj)
        if format:
            df[col] = df[col].dt.strftime(format_str)
    return df




def create_view_from_qs(qs, view_name, materialized=True, ufields=None, owner='postgres', sql_permissions='', read_only_users=None,):
    """Utility function to create a DB view from a django queryset.
    Args:
        qs (django.db.QuerySet): The queryset to be translated into a view
        view_name (str): View name to be created/updated
        materialized (bool): If True will create a materialized view. default True
        ufields (list): List of field names to create a unique index on
        sql_permissions (str, optional): SQL to be executed after the view is created,
            intended for setting permissions but could be any raw SQL.
            i.e. ALTER TABLE public.{view_name} OWNER TO postgres;
        read_only_users (list, optional): List of usernames to grant read-only permisions to. 
            these will be added to the sql_permissions paramter. defaults to []
    """    
    if not materialized and ufields:
        raise AssertionError("Unique indexes only allowed on materialized views")
    if not read_only_users:
        read_only_users = []
    qstr, params = qs.query.sql_with_params()
    vstr = 'MATERIALIZED VIEW' if materialized else 'VIEW'
    drop_qstr1 = f''' DROP VIEW IF EXISTS "{view_name}" '''
    drop_qstr2 = f''' DROP MATERIALIZED VIEW IF EXISTS "{view_name}" '''
    qstr = f''' CREATE {vstr} "{view_name}" AS {qstr} '''
    index_qstr = None
    if ufields:
        index_name = f'unique_{view_name}'
        index_drop = f"DROP INDEX IF EXISTS {index_name}"
        index_qstr = f"CREATE UNIQUE INDEX {index_name} ON {view_name} ({', '.join(ufields)})"
    with connection.cursor() as cursor:
        # Drop existing view (materialized or regular) with the name view_name
        for dstr in [drop_qstr1, drop_qstr2]:
            try:
                cursor.execute(dstr)
            except ProgrammingError:
                pass
        # main view creation
        cursor.execute(qstr, params)
        # unique index creation
        if index_qstr:
            cursor.execute(index_drop)
            cursor.execute(index_qstr)

    std_perms = f'''
        ALTER TABLE public.{view_name} OWNER TO {owner};
        GRANT ALL ON TABLE public.{view_name} TO {owner};
    '''
    with connection.cursor() as cursor:
        cursor.execute(std_perms)

    for ruser in read_only_users:
        sql_permissions += f''' GRANT SELECT ON TABLE public.{view_name} TO {ruser}; '''  

    if sql_permissions:
        with connection.cursor() as cursor:
            cursor.execute(sql_permissions)



from .models import Person
from django_aux.filters import FilterSetBase, MetaBase
from crispy_forms.layout import *


class PersonFilter(FilterSetBase):
    ''' A filter for Person objects '''
    class Meta(MetaBase):
        model = Person
        fields = {
            'last_name': ['icontains'],
            'first_name': ['icontains'],
        }
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.form.helper.layout = Layout(
            Fieldset('',
                Row(
                    Div('last_name__icontains', css_class='ml-2 col-flex'),
                    Div('first_name__icontains', css_class='ml-2 col-flex'),
                ),
            ),
        ) 


# Generated by Django 4.0.6 on 2023-03-03 14:03

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Day',
            fields=[
                ('id', models.BigAutoField(primary_key=True, serialize=False)),
                ('date', models.DateField(unique=True)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='Month',
            fields=[
                ('id', models.BigAutoField(primary_key=True, serialize=False)),
                ('date', models.DateField()),
                ('year_num', models.PositiveSmallIntegerField()),
                ('month_num', models.PositiveSmallIntegerField(choices=[(1, '1'), (2, '2'), (3, '3'), (4, '4'), (5, '5'), (6, '6'), (7, '7'), (8, '8'), (9, '9'), (10, '10'), (11, '11'), (12, '12')])),
                ('name', models.CharField(choices=[('January', 'January'), ('February', 'February'), ('March', 'March'), ('April', 'April'), ('May', 'May'), ('June', 'June'), ('July', 'July'), ('August', 'August'), ('September', 'September'), ('October', 'October'), ('November', 'November'), ('December', 'December')], max_length=10)),
                ('abbr', models.CharField(choices=[('January', 'January'), ('February', 'February'), ('March', 'March'), ('April', 'April'), ('May', 'May'), ('June', 'June'), ('July', 'July'), ('August', 'August'), ('September', 'September'), ('October', 'October'), ('November', 'November'), ('December', 'December')], max_length=10)),
            ],
        ),
        migrations.CreateModel(
            name='Week',
            fields=[
                ('id', models.BigAutoField(primary_key=True, serialize=False)),
                ('date', models.DateField()),
                ('year_num', models.PositiveSmallIntegerField()),
                ('week_num', models.PositiveSmallIntegerField(choices=[(1, '1'), (2, '2'), (3, '3'), (4, '4'), (5, '5'), (6, '6'), (7, '7'), (8, '8'), (9, '9'), (10, '10'), (11, '11'), (12, '12'), (13, '13'), (14, '14'), (15, '15'), (16, '16'), (17, '17'), (18, '18'), (19, '19'), (20, '20'), (21, '21'), (22, '22'), (23, '23'), (24, '24'), (25, '25'), (26, '26'), (27, '27'), (28, '28'), (29, '29'), (30, '30'), (31, '31'), (32, '32'), (33, '33'), (34, '34'), (35, '35'), (36, '36'), (37, '37'), (38, '38'), (39, '39'), (40, '40'), (41, '41'), (42, '42'), (43, '43'), (44, '44'), (45, '45'), (46, '46'), (47, '47'), (48, '48'), (49, '49'), (50, '50'), (51, '51'), (52, '52'), (53, '53')])),
            ],
        ),
        migrations.CreateModel(
            name='Year',
            fields=[
                ('id', models.BigAutoField(primary_key=True, serialize=False)),
                ('date', models.DateField()),
                ('year_num', models.PositiveSmallIntegerField(unique=True)),
                ('is_leap_year', models.BooleanField()),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.AddConstraint(
            model_name='week',
            constraint=models.UniqueConstraint(fields=('year_num', 'week_num'), name='unique_week'),
        ),
        migrations.AddField(
            model_name='month',
            name='year',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, to='django_aux_timeperiods.year'),
        ),
        migrations.AddField(
            model_name='day',
            name='month',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, to='django_aux_timeperiods.month'),
        ),
        migrations.AddField(
            model_name='day',
            name='week',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, to='django_aux_timeperiods.week'),
        ),
        migrations.AddConstraint(
            model_name='month',
            constraint=models.UniqueConstraint(fields=('year_num', 'month_num'), name='unique_month'),
        ),
    ]



import django_tables2 as tables

from .models import Country, Subdivision, Address


class CountryTable(tables.Table):
    class Meta:
        model = Country
        exclude = []

class SubdivisionTable(tables.Table):
    class Meta:
        model = Subdivision
        exclude = []

class AddressTable(tables.Table):
    class Meta:
        model = Address
        exclude = []


from django_aux.forms import ModelFormBase
from .models import Person

class PersonFormWithRequest(ModelFormBase):
    class Meta:
        model = Person
        exclude = []

    def __init__(self, *args, request=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.request = request
        
class PersonForm(ModelFormBase):
    class Meta:
        model = Person
        exclude = []


from django_tables2 import Table
from request.models import Request


class RequestTable(Table):
    ''' Table for displaying user http request instances '''

    class Meta:
        model = Request
        exclude = []
        sequence = []


from django_tables2 import SingleTableMixin
from django.shortcuts import redirect
from django.http import HttpResponseRedirect
from django.urls import reverse_lazy, reverse
from django.db.models import Q, ProtectedError
from pandas import isna, DataFrame as DF, to_datetime
import inspect
from django.contrib import messages
from django.db.models import F, Count
from django_filters.views import FilterView
from django_pandas.io import read_frame
from django.contrib.auth.mixins import UserPassesTestMixin
import plotly.express as px
from plotly import offline
from django.views.generic import DeleteView


class DeleteProtectedView(DeleteView):
    ''' Subclass of django.views.genderic.DeleteView that handles 
    the attempted deltion of Protected FKs '''
    protected_error_msg = "This employee is associated with data and cannot be deleted"

    def get_onclick_cancel_action(self, href=None):
        if href == None:
            href = self.request.session['next']
        if href:
            return f"location.href = '{href}'"
        else:
            return "history.back()"

    def get_protected_error_msg(self):
        ''' Returns the error message to be displayed  '''
        obj = self.get_object()
        return f"{obj} is associted with exisiting data and cannot be deleted"


    def post(self, request, *args, **kwargs):
        #Try super but catch ProtectedError
        try:
            return super().post(request, *args, **kwargs)
        except ProtectedError:
            #When emp FKs to PartRun onDelete is set to PROTECT
            messages.error(request, self.get_protected_error_msg())
            http_referer = request.META.get('HTTP_REFERER')
            if http_referer:
                return HttpResponseRedirect(http_referer)
            else:
                return self.get_success_url()

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context['onclick_cancel_action'] = self.get_onclick_cancel_action()
        return context


class SaveFilterMixinNT:
    """ This Mixin Can be used to save
    the users filter selections after navegating away from the lookup page"""

    def get(self, request, *args, **kwargs):
        view_name = self.__class__.__name__
        qstr = request.GET.urlencode()
        full_path = request.get_full_path()
        if qstr != '':
            self.request.session[f'{view_name}_qstr'] = qstr
            return super().get(request)
        else:
            qstr = request.session.get(f'{view_name}_qstr')
            if qstr != None:
                fp = full_path + '?' + qstr
            else:
                fp = full_path
            if isna(qstr) or '_export' in fp:
                return super().get(request, *args, **kwargs)
            else:
                return redirect(fp)

    def get_filterset_kwargs(self, *args):
        kwargs = super().get_filterset_kwargs(*args)
        if 'clear_filter' in self.request.GET.keys():
            return {}
        else:
            return kwargs

class SaveFilterMixin(SingleTableMixin, SaveFilterMixinNT):
    ''' SaveFilterMixin Classic (for use with SingleTableMixin) '''

class InlineFormsetMixin:
    ''' This mixin allows for multiple formset factories to be injected and processed in a form view '''
    factories = [] # list of dictionaries that must contain the key factory and the value of a formset factory instance, helper and herder are optional
    form_helper = None
    template_name = 'django_aux/inline-formset.html'
    addlines_url = None # set this in the view to have the addlines btn redirect here with self.object.pk as an arg
    
    def get_context_data(self, *args, factories=None, **kwargs):
        context = super().get_context_data(*args, **kwargs)
        if factories == None:
            factories = [fd.copy() for fd in self.factories]
            for fd in factories:
                if self.request.POST:
                    fd['factory'] = fd['factory'](self.request.POST, instance=self.object)
                else:
                    fd['factory'] = fd['factory'](instance=self.object)
        context['factories'] = factories
        context['form_helper']=self.form_helper if self.form_helper else None
        return context
            
    def post(self, request, *args, **kwargs):
        try:
            self.object = self.get_object()
        except AttributeError:
            self.object = None
        context = self.get_context_data()
        form = self.get_form()
        if not form.is_valid():
            return self.form_invalid(form, factories=context['factories'])
        fdsv = True
        for fd in context['factories']:
            if not fd['factory'].is_valid():
                print(fd['factory'].errors)
                fdsv = False
        if fdsv and form.is_valid():
            return self.form_valid(form)
        else:
            return self.form_invalid(form=form, factories=context['factories'])

    def form_valid(self, form):
        self.object = form.save()
        context = self.get_context_data()
        for fd in context['factories']:
            if fd['factory'].is_valid():
                fd['factory'].save()
        self.object.save()
        return super().form_valid(form)

    def form_invalid(self, form, factories):
        return self.render_to_response(self.get_context_data(form=form, factories=factories))

    def get_success_url(self):
        if 'addlines' in self.request.POST and self.addlines_url:
            #user defined a redirect url for the addlines btn, default behavior sends pk as arg
            return reverse_lazy(self.addlines_url, kwargs={'pk':self.object.pk})
        elif 'addlines' in self.request.POST and not self.addlines_url:
            #adding lines to form but no redirect url given. Use default behavior
            return self.request.path_info
        else:
            #add lines was not clicked, use default success_url
            return super().get_success_url()

class DashFilterView(FilterView):
    ''' A view for passing filter qs to a dash app '''
    plotly_app_name = None #set this for dash functionality
    template_name = 'django_aux/dash-filter.html'
    values_args = []    #passing all sorts of args and kwargs for backend qs
    values_kwargs = {}
    annotate_kwargs = {}
    filter_args = []
    filter_kwargs = {}
    to_json_kwargs = dict(date_format='iso', orient='records')

    def get_context_data(self, *args, **kwargs):   
        context = super().get_context_data(*args, **kwargs)
        context['plotly_app_name'] = self.plotly_app_name
        context['style_str'] = "min-width:1500px"        
        context['initial_arguments'] = self.get_initial_arguments()
        return context
        
    def get_initial_arguments(self):
        qs = self.object_list.filter(
            *self.filter_args, **self.filter_kwargs
        ).annotate(
            **self.annotate_kwargs
        ).values(
            *self.values_args, **self.values_kwargs
        )
        df = read_frame(qs).astype(str)
        data = df.to_json(**self.to_json_kwargs)
        return {'data':{'children':data}}

class SaveFormMixin:
    """ This Mixin Can be used with any view that uses a form mixin to
    save a users form selection even if they navigate away from the page"""
    ignore_fields = []

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.view_name = self.__class__

    def get_form(self, **kwargs):
        form = super().get_form(**kwargs)
        if self.request.method == 'GET':
            if len(self.request.GET) > 0:
                qd = self.request.GET
                self.request.session[f'{self.view_name}_form_initial'] = qd
            initial = self.request.session.get(
                f'{self.view_name}_form_initial')
            if initial not in [{}, None]:
                for field in self.ignore_fields:
                    if field in initial.keys():
                        initial.pop(field)
                form = self.form_class(initial=initial)

        if self.request.method == 'POST':
            self.request.session[f'{self.view_name}_form_initial'] = form.data

        return form

class RedirectPrevMixin:
    ''' This mixin will redirect user to the page they came from if 
    form successful OR if "cancel" is in post data  (Uses session data)'''
    redirect_exceptions = [] # list of paths or partial paths that should not be redirected to and the name of the view to use instead
    # i.e.
    # redirect_exceptions = [('person-delete', 'person-lookup')]

    @property
    def form_takes_request_arg(self):
        ''' Property returns True if any of the Parent classes of the views form_class take an 
        argument "request" to their consctructor '''
        form_class = getattr(self.__class__, 'form_class')
        if not form_class: return
        for base in inspect.getmro(form_class):
            if 'request' in inspect.signature(base.__init__).parameters.keys():
                return True
        return False

    def get_form_kwargs(self, *args, **kwargs):
        ''' Method inserts request into the forms kwargs *IF* the forms constructor allows for it '''
        if not self.form_takes_request_arg:
            return super().get_form_kwargs()
        kwargs = super().get_form_kwargs()
        kwargs['request'] = self.request
        return kwargs

    def get_next_is_exception(self, next):
        for pattern, reverse_name  in self.redirect_exceptions:
            if pattern in next:
                return reverse_name  
        return None

    def get(self, request, *args, **kwargs):
        ''' Extends the get method to store where the user was prior to this page '''
        next = request.META.get('HTTP_REFERER')
        if next == None: 
            next = ''
        same_path = request.path in next # redirected from the same page, dont overrwrite next
        exception_reverse_name = self.get_next_is_exception(next)
        if exception_reverse_name:
            request.session['next'] = reverse(exception_reverse_name)
        elif not same_path:
            request.session['next'] = next

        return super().get(request, *args, **kwargs)

    def get_success_url(self):
        ''' If next was stored redirect there, otherwise return super() '''
        next = self.request.session.get('next')
        if next: 
            return next
        return super().get_success_url()

    def post(self, request, *args, **kwargs):
        ''' Override post method to redirect to "next" *IF* 'cancel' is present 
        in POST and next is defined in user session '''
        next = self.request.session.get('next')
        cancel = 'cancel' in self.request.POST.keys()
        if next and cancel:
            return HttpResponseRedirect(next)
        else:
            return super().post(request, *args, **kwargs)

class CheckGroupPermMixin(UserPassesTestMixin):
    ''' This mixin will extend the test_func of any view using
    the UserPassesTestMixin to check the requesting users group membership '''
    allowed_groups = [] #Overwrite, or append, to this with the name of the required django groups
    allow_superusers = True #If true then a django superuser will be granted access regardless of group membership

    def test_func(self):
        ''' an overwritten UserPassesTest test_func that checks group membership '''
        #if allowed groups is an empty list it is taken to be permissive
        if self.allowed_groups == []: return True
        #Now begin the test to see if this user can proceed
        #Test 1: Is the user a super user? (If allowed)
        if self.allow_superusers and self.request.user.is_superuser:
            return True
        #Test 2: Does the user belong to an allowed group?
        user_groups = self.request.user.groups.all().values_list('name',flat=True)
        intersect = set(user_groups).intersection(self.allowed_groups)
        if len(intersect) > 0:
            return True
        return False


class PlotlyMixin:
    ''' This mixin generates a dynamic plotly plot, must be used with a filter that inherits PlotSettingsFilterMixin '''
    plot_width = 1000
    plot_height = None
    max_records = 1_000_000_000


    def check_qs_count(self):
        """ Performs a check to see if the queryset count is greater than the max allowed records

        Returns:
            bool: Indicting whether or the qs was too large
        """        
        if self.object_list.count() > self.max_records:
            emsg = f'Number of records must be < {self.max_records} to produce plot.  Please filter more'
            messages.warning(self.request, emsg)
            return False
        return True   


    @staticmethod
    def add_values(value, vargs, vkwargs):
        """Static method that takes a "value" determines its type and add it to vargs or vkwargs appropriately
        Args:
            value (Any): either a argument or keyword argument to be passed to the Queryset values method
            vargs (list): list of arguments to be passed to Queryset values method
            vkwargs (dict): dict of keyword arguments to be passed to Queryset values method

        """        
        if type(value) == dict:
            vkwargs.update(value)
        else:
            vargs.append(value)

    def get_plot_settings(self):
        """ Method pulls users plot settings from the get request and stores them as instance attributes """ 
        attrs = ['x','y','color','plot_type','aggregate_by','N_min','y_min','y_max']
        for attr in attrs:
            val = self.request.GET.get(attr)
            if val == '': val=None
            setattr(self,attr,val)
        if self.N_min in [None, '']:
            self.N_min = 0
        self.N_min = int(self.N_min)
        ### Get x and y verbose
        self.yd = self.Y_CONFIG.get(self.y)
        self.y_verbose = self.yd.get('verbose') if self.yd else ''
        if self.y_verbose == None:
            self.y_verbose = self.y
        self.x_verbose = dict(self.choices.get('X_CHOICES')).get(self.x)
        if self.x_verbose == None:
            self.x_verbose = self.x

    def get_vargs_vkwargs(self):
        """Method estabishes and returns args and kwargs to be passed to the
            Queryset values method

        Returns:
            tuple: Tuple containing vargs and vkwargs
        """          
        vargs = []
        vkwargs = {} 
        for label in [self.x, self.color, self.aggregate_by]:
            if label == None:
                continue
            val = self.CHOICE_VALUES_MAP.get(label)
            self.add_values(val, vargs, vkwargs) 
        return vargs, vkwargs  

    def get_akwargs(self):
        """Method estabishes and returns the kwargs to be passed to the
            Queryset annotate method

        Returns:
            dict: a dict containing the annotate kwargs
        """          
        agg_func = self.yd.get('agg_expr')
        akwargs = {f'{self.y}':agg_func}
        akwargs['N'] = Count('id')
        return akwargs 

    def get_grouped_qs(self):
        """Method generates and returns the grouped/aggregated queryset to be used in the plot

        Returns:
            Queryset: The grouped queryset to be plotted
        """        
        qs = self.object_list
        vargs, vkwargs = self.get_vargs_vkwargs()
        akwargs = self.get_akwargs()
        fkwargs = {'N__gte':self.N_min}
        if self.y_min: fkwargs[f'{self.y}__gte'] = self.y_min
        if self.y_max: fkwargs[f'{self.y}__lte'] = self.y_max
        fargs = [~Q(**{f'{self.x}':None})]
        if self.color != None: fargs.append(~Q(**{f'{self.color}':None}))
        if self.aggregate_by != None: fargs.append(~Q(**{f'{self.aggregate_by}':None}))
        gqs = qs.values(*vargs, **vkwargs).annotate(**akwargs).filter(*fargs, **fkwargs).order_by(self.x)
        return gqs

    def get_px_args_kwargs_obj(self):
        """Method generates and returns the args, kwargs and plotly.express object to be used in the plot

        Returns:
            tuple: tuple containg the args, kwargs and px object
        """        
        args = [self.plot_df]
        kwargs = dict(x=self.x, y=self.y, width=self.plot_width, height=self.plot_height)
        if self.color:
            kwargs['color'] = self.color
        if 'bar' in self.plot_type:
            if self.plot_type.endswith('g'):
                kwargs['barmode'] = 'group'
            plot_obj = px.bar
        if self.plot_type == 'box':
            plot_obj = px.box
        if self.plot_type == 'line':
            plot_obj = px.line
        if self.plot_type == 'violin':
            kwargs['points'] = 'all'
            kwargs['box'] = True
            plot_obj = px.violin
        if self.plot_type == 'scatter':
            plot_obj = px.scatter
        return args, kwargs, plot_obj

    def get_fig(self):
        """Method generates and returns a plotly Figure object using 
            class variables set in the derived class

        Returns:
            plotly.graph_objects.Figure: A plotly Figure instance
        """    
        self.plot_df = DF()    
        if self.check_qs_count() == False: 
            return
        self.get_plot_settings()
        if self.x == None or self.y == None:
            return None
        gqs = self.get_grouped_qs()
        self.plot_df = read_frame(gqs)
        args, kwargs, plot_obj = self.get_px_args_kwargs_obj()
        fig = plot_obj(*args, **kwargs)
        fig.update_layout(
            xaxis_title=self.x_verbose, yaxis_title=self.y_verbose
        )
        return offline.plot(fig, auto_open=False, output_type="div")


    def get_filterset_kwargs(self, filterset_class):
        kwargs = super().get_filterset_kwargs(filterset_class)
        x_choices = self.X_CHOICES
        agg_choices = [(None,'------')]
        color_choices = [(None,'------')]
        if not hasattr(self, 'COLOR_CHOICES'):
            color_choices += self.X_CHOICES
        else:
            color_choices = self.COLOR_CHOICES
        if not hasattr(self, 'AGG_CHOICES'):
            agg_choices += self.X_CHOICES   
        else:
            agg_choices = self.AGG_CHOICES

        y_choices = []
        for key, yd in self.Y_CONFIG.items():
            y_choices.append((key, yd.get('verbose')))  

        choices = dict(
            X_CHOICES=x_choices,
            Y_CHOICES=y_choices,
            COLOR_CHOICES=color_choices,
            PLOT_TYPE_CHOICES=(
                ('barg', 'Bar-Grouped'), ('bars', 'Bar-Stacked'),
                ('line', 'Line'), ('scatter', 'Scatter'),
                ('box', 'Box'), ('violin', 'Violin'),
            ),
            AGG_CHOICES=agg_choices
        )
        self.choices = choices
        kwargs['choices'] = choices
        return kwargs

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context['fig'] = self.get_fig()
        context['plot_df'] = self.plot_df
        return context


class SinglePlotMixin:
    ''' This mixin creates a potly figure based on a FilterView that is using
    the PlotSettingsFilterMixin '''
    plot_width = 1000
    max_records = 1_000_000_000

    def check_qs_count(self):
        if self.object_list.count() > self.max_records:
            emsg = f'Number of records must be < {self.max_records} to produce plot.  Please filter more'
            messages.warning(self.request, emsg)
            return False
        return True       

    def get_fig(self):
        self.plot_df = DF()
        
        if self.check_qs_count() == False:
            return

        ### Get Plot Settings ###
        x = self.request.GET.get('x')
        y = self.request.GET.get('y')
        color = self.request.GET.get('color')
        plot_type = self.request.GET.get('plot_type')
        agg_by = self.request.GET.get('aggregate_by')
        N_min = self.request.GET.get('N_min')
        if x == None or y == None:
            return None
        if N_min in [None, '']:
            N_min = 0
        N_min = int(N_min)

        ############################
        qs = self.object_list
        qs = qs.annotate(dtg_=F(self.dtg_str))

        keep_vals = ['dtg_']
        if x not in ['day', 'week', 'month', 'quarter', 'year']:
            keep_vals.append(x)
        if agg_by not in ['day', 'week', 'month', 'quarter', 'year'] and agg_by != None:
            keep_vals.append(agg_by)

        yd = self.Y_CONFIG.get(y)
        for val in yd.get('val_list'):
            keep_vals.append(val)

        ### Get x and y verbose
        y_verbose = yd.get('verbose')
        if y_verbose == None:
            y_verbose = y
        x_verbose = dict(self.choices.get('X_CHOICES')).get(x)
        if x_verbose == None:
            x_verbose = x

        if color and color not in keep_vals:
            keep_vals.append(color)
        df = read_frame(qs.values(*keep_vals))
        if df.empty:
            self.plot_df = df
            return
        df.dtg_ = to_datetime(df.dtg_)

        period_dict = {
            'day': {'pcode': 'd', 'date_format': "%Y-%m-%d"},
            'week': {'pcode': 'W', 'date_format': "%Y-W%U"},
            'month': {'pcode': 'M', 'date_format': "%Y-%m"},
            'quarter': {'pcode': 'Q', 'date_format': "%Y-Q%q"},
            'year': {'pcode': 'Y', 'date_format': "%Y"},
        }
        date_format = None
        pcode = None
        for key, d in period_dict.items():
            if agg_by == key or x == key:
                df[key] = df['dtg_'].dt.to_period(
                    d.get('pcode')).dt.to_timestamp()
                if x == key:
                    pcode = d.get('pcode')
                    date_format = d.get('date_format')

        groupby = [agg_by]
        if x != agg_by:
            groupby.append(x)
        if color and color not in agg_by:
            groupby.append(color)

        odf = df.groupby(groupby).apply(
            yd.get('func')).reset_index().rename(columns={0: y})
        ndf = df.groupby(groupby).count().reset_index()
        odf['N'] = ndf.iloc[:, -1]
        ### Filter Data Frame to Only include sample size > N_min
        odf = odf[odf['N'] >= N_min]
        # If filtering makes empty DF set plot_df and return None
        if odf.empty:
            self.plot_df = odf
            return
        ####
        args = [odf]
        kwargs = dict(x=x, y=y, width=self.plot_width)
        if color:
            kwargs['color'] = color
        if 'bar' in plot_type:
            if plot_type.endswith('g'):
                kwargs['barmode'] = 'group'
            plot_obj = px.bar
        if plot_type == 'box':
            plot_obj = px.box
        if plot_type == 'line':
            plot_obj = px.line
        if plot_type == 'violin':
            kwargs['points'] = 'all'
            kwargs['box'] = True
            plot_obj = px.violin
        if plot_type == 'scatter':
            plot_obj = px.scatter

        odf = odf.reset_index(drop=True)
        self.plot_df = odf
        fig = plot_obj(*args, **kwargs)

        if date_format:
            todf = odf.copy()
            if len(todf) > 15:
                c = round(len(todf)/15, -1)
                todf = todf[(todf.index % c == 0) | (
                    todf.index == todf.index.max())]
            tvals = todf[x]
            fig.update_xaxes(
                tickvals=tvals,
                tickformat=date_format,
            )
        fig.update_layout(
            xaxis_title=x_verbose, yaxis_title=y_verbose
        )
        return offline.plot(fig, auto_open=False, output_type="div")

    def get_filterset_kwargs(self, filterset_class):
        kwargs = super().get_filterset_kwargs(filterset_class)
        choices = dict(
            X_CHOICES=[('year', 'Year'), ('quarter', 'Quarter'),
                       ('month', 'Month'), ('week', 'Week'), ('day', 'Day'), ],
            Y_CHOICES=[],
            COLOR_CHOICES=[(None, '--------')],
            PLOT_TYPE_CHOICES=(
                ('barg', 'Bar-Grouped'), ('bars', 'Bar-Stacked'),
                ('line', 'Line'), ('scatter', 'Scatter'),
                ('box', 'Box'), ('violin', 'Violin'),
            ),
            AGG_CHOICES=[
                ('quarter', 'Quarter'), ('month',
                                         'Month'), ('week', 'Week'), ('day', 'Day')
            ]
        )

        for key, yd in self.Y_CONFIG.items():
            choices.get('Y_CHOICES').append((key, yd.get('verbose')))

        if hasattr(self, 'X_CHOICES'):
            choices['X_CHOICES'] = self.X_CHOICES

        if hasattr(self, 'X_CHOICES_EXTRA'):
            for tup in self.X_CHOICES_EXTRA:
                choices.get('X_CHOICES').append(tup)

        if hasattr(self, 'COLOR_CHOICES'):
            choices['COLOR_CHOICES'] = self.COLOR_CHOICES

        if hasattr(self, 'COLOR_CHOICES_EXTRA'):
            for tup in self.COLOR_CHOICES_EXTRA:
                choices.get('COLOR_CHOICES').append(tup)

        if hasattr(self, 'AGG_CHOICES'):
            choices['AGG_CHOICES'] = self.AGG_CHOICES

        if hasattr(self, 'AGG_CHOICES_EXTRA'):
            for tup in self.AGG_CHOICES_EXTRA:
                choices.get('AGG_CHOICES').append(tup)
        self.choices = choices
        kwargs['choices'] = choices
        return kwargs

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context['fig'] = self.get_fig()
        context['plot_df'] = self.plot_df
        return context




from django.urls import path
from .views import *

urlpatterns = [
    path("person-lookup", PersonLookup.as_view(), name="person-lookup"),
    path("person-create-request", PersonCreateWithRequest.as_view(), name="person-create-request"),
    path("person-create", PersonCreate.as_view(), name="person-create"),
]



from django_filters.views import FilterView
from django_aux.views import SaveFilterMixin, RedirectPrevMixin
from django.views.generic import CreateView
from .tables import *
from .filters import *
from .models import *
from .forms import *


class PersonLookup(SaveFilterMixin, FilterView):
    model = Person
    table_class = PersonTable
    filterset_class = PersonFilter
    template_name = "test.html"


class PersonCreate(RedirectPrevMixin, CreateView):
    model = Person
    form_class = PersonForm
    redirect_exceptions = [('person-delete', 'person-lookup')]
    template_name = "test.html"

class PersonCreateWithRequest(RedirectPrevMixin, CreateView):
    model = Person
    form_class = PersonFormWithRequest
    template_name = "test.html"
    





from django_aux.filters import FilterSetBase, MetaBase
from request.models import Request
from crispy_forms.layout import Layout, Fieldset, Row, Div



class RequestFilter(FilterSetBase):
    ''' A filter for Request objects '''
    
    class Meta(MetaBase):
        model = Request
        fields = {
            'response': ['icontains'],
            'method': ['icontains'],
            'path': ['icontains'],
            'is_secure': ['exact'],
            'ip': ['icontains'],
            'user__username': ['icontains'],
            'referer': ['icontains'],
            'time':['gte', 'lte']
        }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.form.fields['time__gte'].label = 'Received after:'
        self.form.fields['time__lte'].label = 'Received before:'
        self.base_layout = Layout(
            Fieldset('',
                Row(
                    Div('response__icontains', css_class='ml-2 col-flex'),
                    Div('method__icontains', css_class='ml-2 col-flex'),
                    Div('path__icontains', css_class='ml-2 col-flex'),
                    Div('is_secure', css_class='ml-2 col-flex'),
                    Div('ip__icontains', css_class='ml-2 col-flex'),
                    Div('user__username__icontains', css_class='ml-2 col-flex'),
                    Div('referer__icontains', css_class='ml-2 col-flex'),
                ),
                Row(
                    Div('time__gte', css_class='ml-2 col-flex'),
                    Div('time__lte', css_class='ml-2 col-flex'),
                ),
            ),
        )
        self.form.helper.layout = self.base_layout


# python imports
import json, uuid
from math import ceil
from pandas import isna, Series, DataFrame as DF

# django imports
from django.utils.html import format_html, mark_safe

# 3rd party django
import django_tables2 as tables
from django_tables2 import A
from django.contrib.contenttypes.models import ContentType





from django_pandas.io import read_frame
from django_aux.utils import df_tz_convert



class FixedTextColumn(tables.Column):
    ''' Mimics behavior of django_tables2.tables.Column but allows for a fixed text to be rendered '''
    def __init__(self, *args, text='', empty_values=None, **kwargs):
        empty_values = empty_values or []
        super().__init__(*args, empty_values=empty_values, **kwargs)
        self.text = text

    def text_value(self, record, value, **kwargs):
        if self.text is None:
            return value
        return self.text(record) if callable(self.text) else self.text

    def value(self, record, value, **kwargs):
        return self.text_value(record=record, value=value, **kwargs)

    def render(self, record, value, **kwargs):
        return self.text_value(record=record, value=value, **kwargs)


class CheckFkColumn(tables.Column):
    ''' A column checks and displays the existence of a FK relationship '''
    def __init__(self, *args, fk_attr=None, present_symbol='✔', absent_symbol='X', **kwargs):
        super().__init__(*args, **kwargs)
        self.fk_attr = fk_attr if fk_attr.endswith('_id') else f'{fk_attr}_id'
        self.psym = present_symbol
        self.asym = absent_symbol

    def render(self, record, value, **kwargs):
        fk_obj = getattr(record, self.fk_attr)
        check = self.asym if fk_obj==None else self.psym
        return f'{value}{check}'

class RoundNumberColumn(tables.Column):
    """ A column that will round a number and add commas for display

    Args:
        *args: arguments to be passed to django-tables2 Column class __init__ method
        money: (bool default=False) Will append a dollar sign $ to front of value if true
        round_to: (int default=2) The number of decimals to round to
        prefix: (str, optional) This string is concatenated to the front of value
        suffix: (str, optional) This string is concatenated to the end of value
        orderable: (bool default=True) Can django-table sort/order this col. Will be set to false if money, suffix or prefix provided
        **kwargs: arguments to be passed to django-tables2 Column class __init__ method
    """    

    def __init__(self, *args, money=False, round_to=2, prefix='', suffix='', **kwargs):
        super().__init__(*args, **kwargs)
        self.money = money
        self.round_to = round_to
        self.prefix = prefix
        self.suffix = suffix
    
    def render(self, value, **kwargs):
        val = round(value, self.round_to)
        if self.round_to <= 0:
            rstr = f'{val:,.0f}'
        else:
            rstr = f'{val:,.{self.round_to}f}'
        if self.money: 
            rstr = '$' + rstr
        return self.prefix + rstr + self.suffix

    def value(self, value, record):
        ''' Overrides default value method (that would just call render on table export) '''
        return value
      
class CollapseColumnMixin:
    def __init__(
        self, 
        *args, 
        label='Show', label_accessor=None, label_extra='', style=None, nowrap=False, 
        empty_values=None, orderable=False,
        **kwargs   
    ):
        empty_values = empty_values or []
        super().__init__(*args, orderable=orderable, empty_values=empty_values, **kwargs)
        self.label = label
        self.label_accessor = label_accessor
        self.label_extra = label_extra
        self.style = style
        self.nowrap = nowrap


    def get_style(self):
        ''' method returns the style to be applied to the collapsable div '''
        if self.style:
            r = f'{self.style}'
        else:
            r = 'max-width: 25vw;'
        if self.nowrap:
            r += 'white-space: nowrap;'
        return f'"{r}"'

    def get_default_label(self, **kwargs):
        return self.label

    def get_label(self, val, record, value, **kwargs):
        """ A method that returns the label to be used for the collapasble div

        Args:
            value (str, optional): The value of the cell in the table (standard arg from django-tables2). Defaults to None.
            record (_type_, optional): The record representing the row of the table (standard arg from django-tables2). Defaults to None.
            val (_type_, optional): the html val to be rendered in the collapasable div. Defaults to None.

        Returns:
            str: The collapsable div label
        """        
        if self.label_accessor:
            rval = A(self.label_accessor).resolve(record)
        else:
            rval = self.get_default_label(val=val, record=record, value=value, **kwargs)
        if value in [None, {}] or val==None:
            return ''     
        elif getattr(self, 'iterable', None):
            if len(value)==0:
                return ''            
        return str(rval) + self.label_extra

    def final_render(self, val, record, value, **kwargs):
        """ The final html to be rendered in the cell of the table

        Args:
            value (str, optional): The value of the cell in the table (standard arg from django-tables2). Defaults to None.
            record (_type_, optional): The record representing the row of the table (standard arg from django-tables2). Defaults to None.
            val (_type_, optional): the html val to be rendered in the collapasable div. Defaults to None.

        Returns:
            str: the html to be rendered in the cell of the table
        """        

        div_id = 'u' + str(uuid.uuid4()).replace('-','')
        label = self.get_label(val=val, record=record, value=value, **kwargs)
        if label != '':
            return format_html(
                '''<a href="#{}" data-toggle="collapse" aria-expanded="false" class="dropdown-toggle">
                    {}
                </a>
                <ul class="collapse list-styled" id="{}">
                    {}
                </ul>
                ''',
                div_id, label, div_id, mark_safe(val)
            )
        else:
            return ''  

class CollapseColumnBase(CollapseColumnMixin, tables.Column):
    """ Sub-class of tables.Column with CollapseColumn functionality """


class CollapseJsonColumn(CollapseColumnMixin, tables.JSONColumn):
    """ Sub-class of JSONColumn with CollapseColumn functionality """

    def render(self, record, value, **kwargs):
        val = super().render(record, value)
        return self.final_render(val=val, record=record, value=value, **kwargs)


class CollapseUrlColumn(CollapseColumnMixin, tables.URLColumn):
    """ Sub-class of URLColumn with CollapseColumn functionality """

    def render(self, record, value, **kwargs):
        val = super().render(record, value, **kwargs)
        val = f'<a href={val}>{self.text}</a>'
        return self.final_render(val=val, record=record, value=value, **kwargs)


class CollapseDictColumn(CollapseColumnBase):
    """Custom django-tables2 column that will render a dictionary in a collapsable div.

    Args:
        *args (iterable, optional): arguments to be passed to django_tables2 Column (See help(django_tables2.Column) for options)
        *kwargs (iterable, optional): keyword arguments to be passed to django_tables2 Column (See help(django_tables2.Column) for options)
        label (str, optional): The string to be displayed on the collapse target. Defaults to "show"
        label_accessor (str, optional): Passed to django-tables2 Accessor object along with the cell value. 
            Result used as collapse target label
        label_extra (str, optional): Extra text that will be appended to the collapse target label
        style (str, optional): css expression that will be passed to the collapasable divs style parameter
        sort_by (str, optional): None, 'key' or 'value' determines what to sort the dictionary by. 
            keys and values will be pandas astype('string'). defaults to None. 
        ascending (bool, optional): bool that sorts keys/values ascending (True) or descending (False). Default = True
        na_position (str, optional): 'first' or 'last' places records with null values up front or at end respectively
        to_html_kwargs (dict, optional): kwargs to be passed to pd.DataFrame.to_html method. 
            defaults to dict(classes = ['table-bordered', 'table-striped', 'table-sm'], index=False, justify='left')   
        to_html_kwargs_extra (dict, optional): kwargs to be added to the to_html_kwargs (typically used for "adding" to defaults). Defaults to {}      
    """    
      
    def __init__(
        self, *args, 
        sort_by = None, 
        ascending = True, 
        na_position = 'last', 
        to_html_kwargs = None, 
        to_html_kwargs_extra = None, 
        **kwargs
    ):                
        super().__init__(*args, **kwargs)
        assert sort_by in [None, 'key', 'value'], 'Invalid sort_by arg. Options are None, key or value'
        self.sort_by = sort_by
        self.ascending = ascending
        self.na_position = na_position
        if to_html_kwargs==None:
            self.to_html_kwargs = dict(
                classes = ['table-bordered', 'table-striped', 'table-sm'],
                index=False, justify='left'
            )                    
        else: 
            self.to_html_kwargs = to_html_kwargs
        self.to_html_kwargs_extra = {} if to_html_kwargs_extra==None else to_html_kwargs_extra
        self.to_html_kwargs.update(self.to_html_kwargs_extra)
        
    def render(self, record, value, **kwargs):
        val = self.get_dictionary_html(record=record, value=value, **kwargs)        
        return self.final_render(value=value, record=record, val=val)

    def value(self, value, **kwargs):
        return value

    def get_dictionary(self, value, **kwargs):
        if isna(value):
            value = {}        
        if type(value) != dict:
            try:
                value = json.loads(value)
            except ValueError:
                value = {}
        return value

    def get_dictionary_html(self, value, **kwargs):
        d = self.get_dictionary(value=value, **kwargs)
        df = DF(Series(d), columns=['value'])
        df = df.reset_index().rename(columns={'index':'key'})
        if self.sort_by:
            df = df.astype(
                {'value':'string','key':'string'}
            ).sort_values(self.sort_by, ascending=self.ascending, na_position=self.na_position)
        df_html = df.to_html(**self.to_html_kwargs)
        return format_html('<div style={}>{}</div>', mark_safe(self.get_style()), mark_safe(df_html))


class CollapseGenericForeignKey(CollapseDictColumn):

    def get_default_label(self, value, **kwargs):
        if not value:
            return
        return f'{value._meta.model.__name__} [{value.pk}]'

    def get_dictionary(self, record, value, **kwargs):
        if not value:
            return 
        model = value._meta.model
        content_type = getattr(record, 'content_type', None) or ContentType.objects.get_for_model(model)
        return {
            'App Label': content_type.app_label,
            'Model Name': model.__name__,
            'Object PK': value.pk,
            'Object': f'{value}'
        } 


class CollapseDataFrameColumn(CollapseColumnBase):
    """Custom django-tables2 column that will render a queryset as a pandas.DataFrame using the 
        pandas.DataFrame.to_html method in a collapsable div.

    Args:
        label (str, optional): text to be used on the collapse link. Defaults to 'Show'.
        group_by (bool, optional): Determines the order that values and annotate methods are applied to the qs. 
            If False: annoate then values. If True: values then annotate. Defaults to False.
        filter_args (list, optional): args passed to qs.filter method. Defaults to [].
        filter_kwargs (dict, optional): kwargs passed to qs.filter method. Defaults to {}.
        annotate_kwargs (dict, optional): kwargs passed to qs.annotate method. Ignored if use_read_frame=True.Defaults to {}.
        values_args (list, optional): args passed to qs.values method. Ignored if use_read_frame=True. Defaults to [].
        values_kwargs (dict, optional): args passed to qs.values method. Ignored if use_read_frame=True. Defaults to {}.
        order_by_args (list, optional): args passed to qs.order_by method. Defaults to [].
        limit (int, optional): limits the qs by slicing it qs[:limit]. Defaults to None.
        use_read_frame (bool, optional): Boolean indicating if django_pandas.io.read_frame
            should be used to convert the qs to a pandas.DataFrame. Defaults to True.
        fieldnames (list, optional): Passed to django_pandas.io.read_frame fieldnames kwarg. 
            ignored if user_read_frame==False. Defaults to None.
        column_names (list, optional): Passed to django_pandas.io.read_frame column_names kwarg. 
            ignored if user_read_frame==False. Defaults to None.
        to_html_kwargs (dict, optional): kwargs to be passed to df.to_html method. 
            Defaults to dict(classes = ['table-bordered', 'table-striped', 'table-sm'], index=False, justify='left').
        to_html_kwargs_extra (dict, optional): kwargs to be added to to_html_kwargs. Defaults to {}.
        timezone (str, optional): Timzone to use on tz-aware datetime columns. Defaults to settings.TIME_ZONE
        datetime_format (bool, optional): Whether to String format datetime columns. Defaults to True
        datetime_format_str (str, optional): What format to use on datetime columns. Defaults to %a %d %b %Y, %I:%M%p 
    """  
    def __init__(
        self, *args, 
        label='Show',
        group_by = False,
        filter_kwargs = None,
        filter_args = None, 
        annotate_kwargs = None, 
        values_kwargs = None, 
        values_args = None, 
        order_by_args = None, 
        limit = None, 
        use_read_frame = True, 
        fieldnames = None, 
        column_names = None,
        to_html_kwargs = None, 
        to_html_kwargs_extra = None, 
        timezone = None, 
        **kwargs   
    ):                
        super().__init__(*args, **kwargs)
        self.label = label
        self.limit = limit
        self.group_by = group_by
        self.timezone = timezone
        self.filter_kwargs = {} if filter_kwargs==None else filter_kwargs
        self.filter_args = [] if filter_args==None else filter_args
        self.annotate_kwargs = {} if annotate_kwargs==None else annotate_kwargs
        self.values_kwargs = {} if values_kwargs==None else values_kwargs
        self.values_args = [] if values_args==None else values_args
        self.order_by_args = [] if order_by_args==None else order_by_args
        self.use_read_frame = use_read_frame
        self.fieldnames = fieldnames
        self.column_names = column_names
        if to_html_kwargs==None:
            self.to_html_kwargs = dict(
                classes = ['table-bordered', 'table-striped', 'table-sm'],
                index=False, justify='left'
            )                    
        else: 
            self.to_html_kwargs = to_html_kwargs
        self.to_html_kwargs_extra = {} if to_html_kwargs_extra==None else to_html_kwargs_extra
        self.to_html_kwargs.update(self.to_html_kwargs_extra)
        self.no_wrap=False
         
    def get_read_frame_kwargs(self, **kwargs):
        """ Reuturns the kwargs to be passed to read_frame function """ 
        kwargs = {}
        if self.fieldnames:
            kwargs['fieldnames'] = self.fieldnames
        if self.column_names:
            kwargs['column_names'] = self.column_names
        return kwargs       

    def get_queryset(self, value, **kwargs):
        ''' method applies user passed kwargs/args to qs methods '''
        qs = value.filter(
            *self.filter_args, **self.filter_kwargs
        )
        if self.use_read_frame == False:
            if self.group_by:
                qs = qs.values(
                    *self.values_args, **self.values_kwargs
                ).annotate(
                    **self.annotate_kwargs
                )
            else:
                qs = qs.annotate(
                    **self.annotate_kwargs
                ).values(
                    *self.values_args, **self.values_kwargs
                )
        qs = qs.order_by(*self.order_by_args)
        return qs if self.limit==None else qs[:self.limit]

    def get_df_final(self, qs, **kwargs):
        ''' Final steps to pd.DF before render or export '''
        if self.use_read_frame:
            df = read_frame(qs, **self.get_read_frame_kwargs())
        else:
            df = DF(qs)
            if self.column_names:
                df.columns = self.column_names
        df = df_tz_convert(df)
        return df

    def get_df_html(self, qs, **kwargs):
        df = self.get_df_final(qs)
        return df.to_html(**self.to_html_kwargs)

    def render(self, record, value, **kwargs):
        qs = self.get_queryset(value=value, **kwargs)
        if qs.count() == 0: 
            val = None
        else:
            val = mark_safe(self.get_df_html(qs, **kwargs))
        return self.final_render(val=val, record=record, value=value, **kwargs)

    def value(self, value, **kwargs):
        ''' Return the value used during table export '''
        qs = self.get_queryset(value)
        if qs.count() == 0:
            return None
        return self.get_df_final(qs).to_dict(orient='records')

class CollapseIterableColumn(CollapseColumnBase):
    """ Custom django-tables2 column that will render an iterable in a collapsable div.

    Args:
        *args (iterable, optional): arguments to be passed to django_tables2 Column (See help(django_tables2.Column) for options)
        order_items_by: (*fields, default None) Expression fed to order_by django method if iterable is a queryset. Should be comma separated fields or other order_by expressions
        hyperlink: (bool, default False) If true, will attempt to linkify the elements of the iterable
        fkwargs: (**kwargs, default None) Lookup/field parameters used in a Django Quereyset filter
        href_attr: (str, default None) Should be the name of an attribute or field that contains the url for linkified values
        str_attr: (str, default None) Name of attribute that will be used for display in lieu of value's __str__()
        **kwargs (iterable, optional): keyword arguments to be passed to django_tables2 Column (See help(django_tables2.Column) for options)
    """    

    def __init__(self, 
        *args, 
        order_items_by = None,
        hyperlink = False,
        fkwargs = None,
        href_attr = None, 
        str_attr = None,  
        **kwargs
    ):
        super().__init__(*args, **kwargs)
        self.hyperlink = hyperlink
        self.href_attr = href_attr
        self.str_attr = str_attr
        self.order_items_by = order_items_by
        self.fkwargs = fkwargs

    def get_href(self, obj, **kwargs):
        ''' Method derives the href value to be used in hyperlinking list items '''
        if self.href_attr == None:
            return obj.get_absolute_url()
        else:
            return getattr(obj, self.href_attr)

    def get_prepped_value(self, value, **kwargs):
        if self.order_items_by and hasattr(value, 'order_by'):
            value = value.order_by(*self.order_items_by)
        if self.fkwargs:
            value = value.filter(**self.fkwargs)
        return value

    def get_final_value(self, value, **kwargs):
        value = self.get_prepped_value(value)
        val = ''
        style = self.get_style()
        for obj in value:
            obj_val = str(obj) if self.str_attr == None else getattr(obj, self.str_attr)
            if self.hyperlink:
                href = self.get_href(obj)
                obj_val = format_html('<a href={}>{}</a>', mark_safe(href), obj_val)
            val = val + format_html('<li style={}>{}</li>', mark_safe(style), obj_val)
        return val

    def render(self, record, value, **kwargs):
        val = self.get_final_value(value=value, **kwargs)
        if not val:
            return ''
        return self.final_render(val=val, record=record, value=value, **kwargs)

    def value(self, value, **kwargs):
        val = self.get_prepped_value(value)
        if self.str_attr:
            alt_values = []
            for obj in val:
                obj_val = getattr(obj, self.str_attr)
                alt_values.append(obj_val)
            return alt_values
        return list(val)

class CollapseNoniterableColumn(CollapseColumnBase):
    """ Custom django-tables2 column that will render an an object in a collapsable div.

    Args:
        *args (iterable, optional): arguments to be passed to django_tables2 Column (See help(django_tables2.Column) for options)
        hyperlink: (bool, default False) If true, will attempt to linkify the elements of the iterable
        href_attr: (str, default None) Should be the name of an attribute or field that contains the url for linkified values
        **kwargs (iterable, optional): keyword arguments to be passed to django_tables2 Column (See help(django_tables2.Column) for options)
    """   

    def __init__(self, *args, hyperlink=False, href_attr=None, orderable=True, **kwargs):
        super().__init__(*args, orderable=orderable, **kwargs)
        self.hyperlink = hyperlink
        self.href_attr = href_attr

    def get_href(self, obj, **kwargs):
        ''' Method derives the href value to be used in hyperlinking list items '''
        if self.href_attr == None:
            return obj.get_absolute_url()
        else:
            return getattr(obj, self.href_attr)

    def get_prepped_value(self, record, value, **kwargs):
        if self.hyperlink:
            href = self.get_href(record)
            val = format_html('<a href={}>{}</a>', mark_safe(href), value)
        else:
            val = value
        return format_html('<div style={}>{}</div>', mark_safe(self.get_style()), val)

    def render(self, record, value, **kwargs):
        val = self.get_prepped_value(value=value, record=record)
        return self.final_render(val=val, record=record, value=value, **kwargs)   

    def value(self, value, **kwargs):
        return value




class CollapseColumn(CollapseColumnBase):
    ''' Column is meant for columns that have lots of data in each cell to make viewing cleaner'''

    def __init__(
        self, *args, hyperlink=False, href_attr=None,
        iterable=False, str_attr=None, order_by=None, fkwargs=None, property_attr=None, dictionary=False,
        **kwargs
    ):  # Note on kwargs: lavel_accessor used to make dynamic labels, label_extra is a str that adds on to the returned value
        super().__init__(*args, **kwargs)
        self.hyperlink = hyperlink  # Attempts to linkify the elements of an iterable
        # the attribute name to be used to pull the href value if None provided get_absolute_url will be called
        self.href_attr = href_attr
        self.iterable = iterable
        self.str_attr = str_attr
        self.order_by = order_by
        self.fkwargs = fkwargs
        self.property_attr = property_attr
        self.dictionary = dictionary

    def get_href(self, obj):
        ''' Method derives the href value to be used in hyperlinking list items '''
        if self.href_attr == None:
            return obj.get_absolute_url()
        else:
            return getattr(obj, self.href_attr)

    def render(self, value, record):
        if self.property_attr:
            value = getattr(record, self.property_attr)
        if self.dictionary:
            val = self.get_dictionary_val(value=value)        
        elif self.iterable == False:
            val = self.get_noniterable_val(value=value, record=record)
        else: 
            val = self.get_iterable_val(value)
        return self.final_render(value=value, record=record, val=val)
    
    def get_iterable_val(self, value):
        if self.order_by:
            value = value.order_by(self.order_by)
        if self.fkwargs:
            value = value.filter(**self.fkwargs)
        val = ''
        style = self.get_style()
        for obj in value:
            obj_val = str(obj) if self.str_attr == None else getattr(obj, self.str_attr)
            if self.hyperlink:
                href = self.get_href(obj)
                obj_val = f'<a href={href}>{obj_val}</a>'
            val = val + f'<li style={style}>{obj_val}</li>'
        return val

    def get_noniterable_val(self, value, record):
        if self.hyperlink:
            href = self.get_href(record)
            val = f'<a href={href}>{value}</a>'
        else:
            val = value
        return f'<div style={self.get_style()}>{val}</div>'

    def get_dictionary_val(self, value):
        if isna(value):
            value = {}        
        if type(value) != dict:
            value = json.loads(value)
        df = DF(Series(value), columns=['value'])
        df = df.reset_index().rename(columns={'index':'key'})
        df_html = df.to_html(
            classes = ['table-bordered', 'table-striped', 'table-sm'],
            index=False, justify='left', header=False
        )
        return f'<div style={self.get_style()}>{df_html}</div>'


def get_background(value, record, table, bound_column):
    val = (str(value).split('/')[0]).replace(',','')
    val = float(val)
    vals = [getattr(row, bound_column.name) for row in table.data]
    ser = Series(vals).fillna(0)
    max = ser.max()
    if max == 0:
        w = 0
    else:
        w = 100*val/max
    style_str = f'background:linear-gradient(90deg, {bound_column.column.color} {w}%, transparent {w}%)'
    return style_str

class BarChartColumn(tables.Column):
    def __init__(
        self, *args,
        round_to=0,
        color='lightgrey',
        goal_attr=None,
        **kwargs
    ):
        super().__init__(*args, **kwargs)
        self.round_to = round_to
        self.color = color
        self.goal_attr = goal_attr
        td_dict = self.attrs.get('td', {})
        td_dict['style'] = get_background
        self.attrs['td'] = td_dict

    def render(self, value, record):
        if value==None:
            return
        value = round(value, self.round_to)
        if self.round_to > 0:
            rval = f'{value:,}'
        else:
            rval = f'{value:,.0f}'

        if self.goal_attr:
            goal = getattr(record, self.goal_attr, '')
            goal = ceil(goal)
            rval = str(rval) + f'/{goal}'
        return rval

    def value(self, value):
        value = str(value).split('/')[0]
        if not str(value).isnumeric():
            value = 0
        return value

class LastChangeDateColumn(tables.Column):
    ''' This Column can be used with tables that have a model defined and 
    are using django simple-history to track changes'''

    def render(self, record):
        if not hasattr(record, 'history'):
            return None
        else:
            return record.history.first().history_date


class LastChangeUserColumn(tables.Column):
    ''' This Column can be used with tables that have a model defined and 
    are using django simple-history to track changes'''

    def render(self, record):
        if not hasattr(record, 'history'):
            return None
        else:
            user = record.history.first().history_user
            if user == None:
                return 'Automated'
            return user.employee

class LastChangeTypeColumn(tables.Column):
    ''' This Column can be used with tables that have a model defined and 
    are using django simple-history to track changes'''

    def render(self, record):
        if not hasattr(record, 'history'):
            return None
        else:
            return record.history.first().history_type



from django.db import models
import uuid
import names
import random



class Person(models.Model):
    ''' Instance of this model represents a human being '''
    uuid = models.UUIDField(default = uuid.uuid4, editable=False)
    last_name = models.CharField(max_length=100, null=True)
    first_name = models.CharField(max_length=100, null=True)
    middle_name = models.CharField(max_length=100, null=True, blank=True)
    salary = models.FloatField(null=True)

    @property
    def attr_dict(self):
        d = self.__dict__.copy()
        d.pop('_state')
        return d

    @property
    def full_name(self):
        mn = '' if self.middle_name==None else self.middle_name
        return f'{self.first_name} {mn} {self.last_name}'

    @property
    def long_text(self):
        import lipsum
        return lipsum.generate_paragraphs(3)

    @classmethod
    def generate_random_people(cls, wipe_table=False, n=10):
        ''' classmethod to generate random people by calling generate_random_person module '''
        if wipe_table==True:
            cls.objects.all().delete()
        for i in range(n):
            cls.generate_random_person()

    @classmethod
    def generate_random_person(cls, gender=None):
        ''' classmethod to generate a random person '''
        if gender!=None:
            assert gender in ['male', 'female'], 'gender must be either "male" or "female"'
        else:
            gender = random.choice(['female', 'male'])
        name = names.get_full_name(gender=gender)
        fn = name.split(' ')[0]
        mn = names.get_first_name(gender=gender)
        ln = name.split(' ')[1]
        obj, c = cls.objects.get_or_create(last_name=ln, first_name=fn, middle_name=mn)
        return obj

    def get_salary(self):
        ''' Method to get a random salary if none '''
        if self.salary: return
        self.salary = round(random.uniform(15_000.00, 200_000.00), 2)


    def save(self, *args, **kwargs):
        self.get_salary()
        super().save(*args, **kwargs)

    def __str__(self):
        return self.full_name



from django.db import models
import datetime as dt
from django.db.models import UniqueConstraint
import pandas as pd
from zoneinfo import ZoneInfo



class TimePeriodBase(models.Model):
    ''' Base class to hold common methods for TimePeriod models '''
    id = models.BigAutoField(primary_key=True)
    freq_map = {'Year':'Y', 'Month':'M', 'Week':'W', 'Day':'D'}

    class Meta:
        abstract = True

    @property
    def period(self):
        ''' property returns a pandas period object representing this instance '''
        cname = self.__class__.__name__
        if cname == 'Year':
            return pd.Period(year=self.year_num, freq='Y')
        elif cname == 'Month':
            return pd.Period(year=self.year_num, month=self.month_num, freq='M')
        elif cname == 'Week':
            edate = self.end_dates[self.week_num-1]
            return pd.Period(edate, freq='W')
        elif cname == 'Day':
            return pd.Period(self.date, freq='D')

    @classmethod
    def get_or_create_from_date(cls, date):
        ''' Classmethod will get or create a timeperiod object based on the passed date '''
        cname = cls.__name__
        sdtg = pd.Period(date, freq=cls.freq_map[cname]).start_time
        if cname == 'Year':
            return cls.objects.get_or_create(year_num=sdtg.year)
        elif cname == 'Month':
            return cls.objects.get_or_create(year_num=sdtg.year, month_num=sdtg.month)
        elif cname == 'Week':
            return cls.objects.get_or_create(year_num=sdtg.year, week_num=sdtg.week)
        elif cname == 'Day':
            return cls.objects.get_or_create(date=date)

    @classmethod
    def get_or_create_current_period(cls, tzinfo=None):
        ''' Classmethod will create a timeperiod object that encompasses today '''
        now = dt.datetime.now()
        if tzinfo:
            now = now.astimezone(tzinfo)
        return cls.get_or_create_from_date(now.date())

    @classmethod
    def get_or_create_n_from_current(cls, n=0, tzinfo=None):
        ''' Retreives a TimePeriod object n periods from current period '''
        now = dt.datetime.now()
        if tzinfo:
            now = now.astimezone(tzinfo)
        cperiod = pd.Period(now, freq=cls.freq_map[cls.__name__])
        if n==0:
            return cls.get_or_create_from_date(cperiod.start_time.date())
        elif n < 0:
            kwargs = {'end':cperiod}
            i = 0
        elif n > 0:
            kwargs = {'start':cperiod}
            i = -1
        period = pd.period_range(**kwargs, periods=abs(n)+1)[i]
        return cls.get_or_create_from_date(period.start_time.date())


    def get_rel_status(self, dtg=None):
        ''' Method returns the status of a TimePeriod (past, present, future) relative to the passed datetime '''
        if dtg==None:
            dtg = dt.datetime.now()
        tz = dtg.tzinfo
        sdtg = self.period.start_time
        edtg = self.period.end_time
        if tz:
            sdtg = sdtg.tz_localize(tz)
            edtg = edtg.tz_localize(tz)
        if dtg < sdtg:
            return 'future'
        elif sdtg <= dtg <= edtg:
            return 'present'
        else:
            return 'past'


class Year(TimePeriodBase):
    ''' An instance of this model represents a year in the Gregorian Calendar.
        Instances should be initialized with year_num.  Save method 
        will derive the other attributes
    '''
    date = models.DateField()
    year_num = models.PositiveSmallIntegerField(unique=True)
    is_leap_year = models.BooleanField()

    def set_date(self):
        ''' Method to set the date attribute '''
        self.date = dt.date(self.year_num, 1, 1)

    def set_leap_year(self):
        ''' Method to set the leap_year boolean '''
        self.is_leap_year = self.period.is_leap_year

    def save(self, *args, **kwargs):
        self.set_date()
        self.set_leap_year()
        return super().save(*args, **kwargs)

    def __str__(self):
        return str(self.period)

class Month(TimePeriodBase):
    ''' An instance of this model represents a month in the Gregorian Calendar.
        Instances should be initialized with year_num and month_num.  Save method 
        will derive the other attributes
    '''
    NAME_CHOICES = (
        ('January', 'January'), ('February', 'February'), ('March', 'March'),
        ('April', 'April'), ('May', 'May'), ('June', 'June'), ('July', 'July'),
        ('August', 'August'), ('September', 'September'), ('October', 'October'),
        ('November', 'November'), ('December', 'December'))
    ABBR_CHOICES = (
        ('Jan', 'Jan'),('Feb', 'Feb'),('Mar', 'Mar'),('Apr', 'Apr'),('May', 'May'),
        ('Jun', 'Jun'),('Jul', 'Jul'),('Aug', 'Aug'),('Sep', 'Sep'),('Oct', 'Oct'),
        ('Nov', 'Nov'),('Dec', 'Dec')
    )
    MONTH_NUM_CHOICES = [(i,str(i)) for i in range(1,13)]
    date = models.DateField()
    year = models.ForeignKey('Year', on_delete=models.SET_NULL, null=True)
    year_num = models.PositiveSmallIntegerField()
    month_num = models.PositiveSmallIntegerField(choices=MONTH_NUM_CHOICES)
    name = models.CharField(max_length=10, choices=NAME_CHOICES)
    abbr = models.CharField(max_length=10, choices=NAME_CHOICES)

    class Meta:
        constraints = [UniqueConstraint(fields=['year_num', 'month_num'], name='unique_month')]


    def set_date(self):
        ''' Method to derive the date attribute '''
        self.date = dt.date(self.year_num, self.month_num, 1)

    def set_name(self):
        ''' Method sets the name attribute '''
        self.name = self.NAME_CHOICES[self.month_num-1][0]

    def set_abbr(self):
        ''' Method sets the abbr attribute '''
        self.name = self.ABBR_CHOICES[self.month_num-1][0]      

    def set_year(self):
        ''' Method sets the year FK attribute. Creates the instance if necessary '''
        self.year, _ = Year.objects.get_or_create(year_num=self.year_num)

    def save(self, *args, **kwargs):
        self.set_date()
        self.set_name()
        self.set_abbr()
        self.set_year()
        return super().save(*args, **kwargs)

    def __str__(self):
        return str(self.period)

class Week(TimePeriodBase):
    ''' An instance of this model represents a week in the Gregorian Calendar.
        Numbering convention follows ISO 8601. Weeks always start on Monday and 
        week 1 of the year is the first week this the majority of its 7 days in January.
        Instances should be initialized with year_num and week_num.  Save method 
        will derive the other attributes
    '''
    WEEK_NUM_CHOICES = [(i,str(i)) for i in range(1,54)]
    date = models.DateField()
    year_num = models.PositiveSmallIntegerField()
    week_num = models.PositiveSmallIntegerField(choices=WEEK_NUM_CHOICES)

    class Meta:
        constraints = [UniqueConstraint(fields=['year_num', 'week_num'], name='unique_week')]

    @property
    def end_dates(self):
        ''' property returns a list of week end dates for the year of this instance '''
        start = pd.to_datetime(dt.date(self.year_num,1,1)).to_period('W').start_time
        end = pd.to_datetime(dt.date(self.year_num,12,31)).to_period('W').end_time
        dates = pd.date_range(start, end, freq='W')
        if start.week != 1:
            dates = dates[1:]
        if end.week < 52:
            dates = dates[:-1]
        return dates

    def set_date(self):
        ''' Method to derive the date attribute '''
        self.date = self.period.start_time.date()

    def set_year(self):
        ''' Method sets the year FK attribute. Creates the instance if necessary '''
        self.year, _ = Year.objects.get_or_create(year_num=self.year_num)

    def save(self, *args, **kwargs):
        self.set_date()
        self.set_year()
        return super().save(*args, **kwargs)

    def __str__(self):
        return f'{self.year_num} W{str(self.week_num).zfill(2)}'


class Day(TimePeriodBase):
    ''' An instance of this model represents a day in the Gregorian Calendar.
        Instances should be initialized with a date.  Save method 
        will derive the other attributes
    '''
    month = models.ForeignKey('Month', on_delete=models.SET_NULL, null=True)
    week = models.ForeignKey('Week', on_delete=models.SET_NULL, null=True)
    date = models.DateField(unique=True)


    def set_month(self):
        ''' Method sets the year FK attribute. Creates the instance if necessary '''
        self.month, _ = Month.objects.get_or_create(year_num=self.date.year, month_num=self.date.month)

    def set_week(self):
        ''' Method sets the week FK attribute. Creates the instance if necessary '''
        self.week, _ = Week.get_or_create_from_date(self.date)

    def save(self, *args, **kwargs):
        self.set_month()
        self.set_week()
        return super().save(*args, **kwargs)

    def __str__(self):
        return str(self.period)


import django_tables2 as tables
from .models import *


class PersonTable(tables.Table):
    ''' Table for displaying fixed Schedule instances '''
    class Meta:
        model = Person
        exclude = []
        sequence = []